#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import itertools
import operator


class WorldAnchorage2D:

    # constant defs
    C = CENTER = "c"
    E = EAST = "e"
    N = NORTH = "n"
    S = SOUTH = "s"
    W = WEST = "w"

    LEGAL_ANCHORS = (C, E, N, N+E, N+W, S, S+E, S+W, W)

    DEFAULT_ANCHOR = CENTER
    DEFAULT_ORIGIN = None
    DEFAULT_UNIT_VECTOR = (1, 1)

    def __init__ (self, **options):
        self.options = options
        self.anchor = options.get("anchor")
        self.origin = options.get("origin")
        self.unit_vector = options.get("unit_vector")
    # end def

    @property
    def anchor (self):
        return self.__anchor
    # end def

    @anchor.setter
    def anchor (self, value):
        value = str(value or self.DEFAULT_ANCHOR).lower()
        if value in self.LEGAL_ANCHORS:
            self.__anchor = value
        else:
            raise ValueError(
                "anchor should be one of {}."
                .format(self.LEGAL_ANCHORS)
            )
        # end if
    # end def

    @property
    def origin (self):
        """Anchorage origin point for relative coords calculations."""
        return self.__origin
    # end def

    @origin.setter
    def origin (self, value):
        self.__origin = self.point_type(value or self.DEFAULT_ORIGIN)
    # end def

    @property
    def point_type (self):
        """read-only attribute"""
        return WorldPoint2D
    # end def

    @property
    def unit_vector (self):
        return self.__unit_vector
    # end def

    @unit_vector.setter
    def unit_vector (self, value):
        self.__unit_vector = self.vector_type(
            value or self.DEFAULT_UNIT_VECTOR
        )
    # end def

    @property
    def vector_type (self):
        """read-only attribute"""
        return WorldVector2D
    # end def

# end class


class WorldAnchorage3D (WorldAnchorage2D):

    # constant defs
    F = FRONT = "f"
    M = MIDDLE = "m"
    R = REAR = "r"

    LEGAL_ANCHORS = tuple(
        a + b
        for a in (F, M, R)
        for b in WorldAnchorage2D.LEGAL_ANCHORS
    )

    DEFAULT_ANCHOR = MIDDLE + WorldAnchorage2D.CENTER
    DEFAULT_ORIGIN = None
    DEFAULT_UNIT_VECTOR = (1, 1, 1)

    @property
    def point_type (self):
        """read-only attribute"""
        return WorldPoint3D
    # end def

    @property
    def vector_type (self):
        """read-only attribute"""
        return WorldVector3D
    # end def

# end class


class WorldBoxBase2D:

    def __getitem__ (self, key):
        return self.to_tuple()[key]
    # end def

    def __init__ (self, width=0, height=0, **options):
        try:
            (self.width, self.height) = width
        except:
            (self.width, self.height) = (width, height)
        # end try
        self.options = options
    # end def

    def __repr__ (self):
        return str(self.to_tuple())
    # end def

    @property
    def center_x (self):
        """read-only attribute"""
        return self.width//2
    # end def

    @property
    def center_xy (self):
        """read-only attribute"""
        return (self.center_x, self.center_y)
    # end def

    @property
    def center_y (self):
        """read-only attribute"""
        return self.height//2
    # end def

    @property
    def height (self):
        return self.__height
    # end def

    @height.setter
    def height (self, value):
        self.__height = self.normalize_dimension(value)
    # end def

    def normalize_dimension (self, value):
        if value:
            if isinstance(value, (int, float)):
                return abs(value)
            else:
                raise ValueError(
                    "Value must be of int or float type, not '{}'."
                    .format(value)
                )
            # end if
        # end if
        return 0
    # end def

    @property
    def size (self):
        """read-only attribute"""
        return (self.width, self.height)
    # end def

    def to_dict (self):
        return dict(width=self.width, height=self.height)
    # end def

    def to_tuple (self):
        return (self.width, self.height)
    # end def

    @property
    def width (self):
        return self.__width
    # end def

    @width.setter
    def width (self, value):
        self.__width = self.normalize_dimension(value)
    # end def

# end class


class WorldBoxBase3D (WorldBoxBase2D):

    def __init__ (self, width=0, height=0, depth=0, **options):
        try:
            (self.width, self.height, self.depth) = width
        except:
            self.width = width
            self.height = height
            self.depth = depth
        # end try
        self.options = options
    # end def

    def center (self, query=None):
        query = str(query or "xyz").lower()
        _result = []
        if "x" in query:
            _result.append(self.center_x)
        # end if
        if "y" in query:
            _result.append(self.center_y)
        # end if
        if "z" in query:
            _result.append(self.center_z)
        # end if
        return tuple(_result)
    # end def

    @property
    def center_xyz (self):
        """read-only attribute"""
        return (self.center_x, self.center_y, self.center_z)
    # end def

    @property
    def center_z (self):
        """read-only attribute"""
        return self.depth//2
    # end def

    @property
    def depth (self):
        return self.__depth
    # end def

    @depth.setter
    def depth (self, value):
        self.__depth = self.normalize_dimension(value)
    # end def

    @property
    def size (self):
        """read-only attribute"""
        return (self.width, self.height, self.depth)
    # end def

    def to_dict (self):
        return dict(
            width=self.width, height=self.height, depth=self.depth
        )
    # end def

    def to_tuple (self):
        return (self.width, self.height, self.depth)
    # end def

# end class


class WorldArea2D (WorldBoxBase2D):

    def __init__ (self, width=0, height=0, **options):
        super().__init__(width, height, **options)
        self.bounds = options.get("bounds") or options.get("limits")
    # end def

    @property
    def bounds (self):
        return self.__bounds
    # end def

    @bounds.setter
    def bounds (self, value):
        self.__bounds = self.bounds_type(
            *(value or self.default_bounds)
        )
    # end def

    @property
    def bounds_type (self):
        """read-only attribute"""
        return WorldBoundingBox2D
    # end def

    @property
    def default_bounds (self):
        """read-only attribute"""
        return ((0,) * len(self.size), self.size)
    # end def

# end class


class WorldArea3D (WorldBoxBase3D, WorldArea2D):

    def __init__ (self, width=0, height=0, depth=0, **options):
        WorldBoxBase3D.__init__(self, width, height, depth, **options)
        self.bounds = options.get("bounds") or options.get("limits")
    # end def

    @property
    def bounds_type (self):
        """read-only attribute"""
        return WorldBoundingBox3D
    # end def

# end class


class WorldBoundingBox2D:

    def __getitem__ (self, key):
        return self.to_tuple()[key]
    # end def

    def __init__ (self, *points, **options):
        self.reset_bounds(*points)
        self.options = options
    # end def

    def __repr__ (self):
        return str(self.to_tuple())
    # end def

    def center (self, query=None):
        query = str(query or "xy").lower()
        _result = []
        if "x" in query:
            _result.append(self.center_x)
        # end if
        if "y" in query:
            _result.append(self.center_y)
        # end if
        return tuple(_result)
    # end def

    @property
    def center_x (self):
        """read-only attribute"""
        return (self.point1.x + self.point2.x)//2
    # end def

    @property
    def center_xy (self):
        """read-only attribute"""
        return (self.center_x, self.center_y)
    # end def

    @property
    def center_y (self):
        """read-only attribute"""
        return (self.point1.y + self.point2.y)//2
    # end def

    def collided (self, *points):
        return any(map(self.in_bounds, points))
    # end def

    @property
    def corners (self):
        """read-only attribute"""
        return tuple(
            self.point_type(*coords)
            for coords in
                itertools.product(*zip(self.point1, self.point2))
        )
    # end def

    def entered (self, *points):
        _bounds = []
        (_all_x, _all_y) = zip(*points)
        (_min_px, _min_py) = (min(_all_x), min(_all_y))
        (_max_px, _max_py) = (max(_all_x), max(_all_y))
        (_b_x, _b_y) = zip(self.point1, self.point2)
        (_min_bx, _min_by) = (min(_b_x), min(_b_y))
        (_max_bx, _max_by) = (max(_b_x), max(_b_y))
        _min_px_in = bool(_min_bx <= _min_px <= _max_bx)
        _max_px_in = bool(_min_bx <= _max_px <= _max_bx)
        _min_py_in = bool(_min_by <= _min_py <= _max_by)
        _max_py_in = bool(_min_by <= _max_py <= _max_by)
        if any((_min_py_in, _max_py_in)):
            if _max_px_in:
                _bounds.append("left")
            # end if
            if _min_px_in:
                _bounds.append("right")
            # end if
        # end if
        if any((_min_px_in, _max_px_in)):
            if _min_py_in:
                _bounds.append("top")
            # end if
            if _max_py_in:
                _bounds.append("bottom")
            # end if
        # end if
        return tuple(_bounds)
    # end def

    def exited (self, *points):
        _bounds = []
        (_all_x, _all_y) = zip(*points)
        (_b_x, _b_y) = zip(self.point1, self.point2)
        if min(_all_x) <= min(_b_x):
            _bounds.append("left")
        # end if
        if max(_all_x) >= max(_b_x):
            _bounds.append("right")
        # end if
        if max(_all_y) >= max(_b_y):
            _bounds.append("top")
        # end if
        if min(_all_y) <= min(_b_y):
            _bounds.append("bottom")
        # end if
        return tuple(_bounds)
    # end def

    @property
    def height (self):
        """read-only attribute"""
        return abs(self.point1.y - self.point2.y)
    # end def

    def in_bounds (self, point):
        point = self.point_type(*point)
        return all(
            map(
                lambda a, b: min(a) <= b <= max(a),
                zip(self.point1, self.point2), point
            )
        )
    # end def

    @property
    def point1 (self):
        return self.__point1
    # end def

    @point1.setter
    def point1 (self, value):
        self.__point1 = self.point_type(*value)
    # end def

    @property
    def point2 (self):
        return self.__point2
    # end def

    @point2.setter
    def point2 (self, value):
        self.__point2 = self.point_type(*value)
    # end def

    @property
    def point_type (self):
        """read-only attribute"""
        return WorldPoint2D
    # end def

    def reset_bounds (self, *points):
        _coords = tuple(zip(*filter(None, points)))
        self.point1 = map(min, _coords)
        self.point2 = map(max, _coords)
    # end def

    @property
    def size (self):
        """read-only attribute"""
        return (self.width, self.height)
    # end def

    def to_dict (self):
        return dict(point1=self.point1, point2=self.point2)
    # end def

    def to_flat_tuple (self):
        return tuple(dim for point in self.to_tuple() for dim in point)
    # end def

    def to_tuple (self):
        return (self.point1, self.point2)
    # end def

    @property
    def width (self):
        """read-only attribute"""
        return abs(self.point1.x - self.point2.x)
    # end def

# end class


class WorldBoundingBox3D (WorldBoundingBox2D):

    def center (self, query=None):
        query = str(query or "xyz").lower()
        _result = []
        if "x" in query:
            _result.append(self.center_x)
        # end if
        if "y" in query:
            _result.append(self.center_y)
        # end if
        if "z" in query:
            _result.append(self.center_z)
        # end if
        return tuple(_result)
    # end def

    @property
    def center_xyz (self):
        """read-only attribute"""
        return (self.center_x, self.center_y, self.center_z)
    # end def

    @property
    def center_z (self):
        """read-only attribute"""
        return (self.point1.z + self.point2.z)//2
    # end def

    @property
    def depth (self):
        """read-only attribute"""
        return abs(self.point1.z - self.point2.z)
    # end def

    def entered (self, *points):
        _bounds = []
        (_all_x, _all_y, _all_z) = zip(*points)
        (_min_px, _min_py, _min_pz) = (
            min(_all_x), min(_all_y), min(_all_z)
        )
        (_max_px, _max_py, _max_pz) = (
            max(_all_x), max(_all_y), max(_all_z)
        )
        (_b_x, _b_y, _b_z) = zip(self.point1, self.point2)
        (_min_bx, _min_by, _min_bz) = (min(_b_x), min(_b_y), min(_b_z))
        (_max_bx, _max_by, _max_bz) = (max(_b_x), max(_b_y), max(_b_z))
        _min_px_in = bool(_min_bx <= _min_px <= _max_bx)
        _max_px_in = bool(_min_bx <= _max_px <= _max_bx)
        _min_py_in = bool(_min_by <= _min_py <= _max_by)
        _max_py_in = bool(_min_by <= _max_py <= _max_by)
        _min_pz_in = bool(_min_bz <= _min_pz <= _max_bz)
        _max_pz_in = bool(_min_bz <= _max_pz <= _max_bz)
        if any((_min_py_in, _max_py_in, _min_pz_in, _max_pz_in)):
            if _max_px_in:
                _bounds.append("left")
            # end if
            if _min_px_in:
                _bounds.append("right")
            # end if
        # end if
        if any((_min_px_in, _max_px_in, _min_pz_in, _max_pz_in)):
            if _min_py_in:
                _bounds.append("top")
            # end if
            if _max_py_in:
                _bounds.append("bottom")
            # end if
        # end if
        if any((_min_px_in, _max_px_in, _min_py_in, _max_py_in)):
            if _max_pz_in:
                _bounds.append("front")
            # end if
            if _min_pz_in:
                _bounds.append("rear")
            # end if
        # end if
        return tuple(_bounds)
    # end def

    def exited (self, *points):
        _bounds = []
        (_all_x, _all_y, _all_z) = zip(*points)
        (_b_x, _b_y, _b_z) = zip(self.point1, self.point2)
        if min(_all_x) <= min(_b_x):
            _bounds.append("left")
        # end if
        if max(_all_x) >= max(_b_x):
            _bounds.append("right")
        # end if
        if max(_all_y) >= max(_b_y):
            _bounds.append("top")
        # end if
        if min(_all_y) <= min(_b_y):
            _bounds.append("bottom")
        # end if
        if min(_all_z) <= min(_b_z):
            _bounds.append("front")
        # end if
        if max(_all_z) >= max(_b_z):
            _bounds.append("rear")
        # end if
        return tuple(_bounds)
    # end def

    @property
    def point_type (self):
        """read-only attribute"""
        return WorldPoint3D
    # end def

    @property
    def size (self):
        """read-only attribute"""
        return (self.width, self.height, self.depth)
    # end def

# end class


class WorldPoint2D:

    def __add__ (self, other):
        try:
            return self.__class__(*map(operator.add, self, other))
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def __floordiv__ (self, other):
        if isinstance(other, (int, float)):
            return self.__class__(*(dim // other for dim in self))
        else:
            return NotImplemented
        # end if
    # end def

    def __getitem__ (self, key):
        return self.to_tuple()[key]
    # end def

    def __init__ (self, x=0, y=0):
        try:
            (self.x, self.y) = x
        except:
            (self.x, self.y) = (x, y)
        # end try
    # end def

    def __mul__ (self, other):
        if isinstance(other, (int, float)):
            return self.__class__(*(dim * other for dim in self))
        else:
            return NotImplemented
        # end if
    # end def

    def __repr__ (self):
        return str(self.to_tuple())
    # end def

    def __rmul__ (self, other):
        return self.__mul__(other)
    # end def

    def __sub__ (self, other):
        try:
            return self.__class__(*map(operator.sub, self, other))
        except TypeError:
            return NotImplemented
        # end try
    # end def

    def __truediv__ (self, other):
        if isinstance(other, (int, float)):
            return self.__class__(*(dim / other for dim in self))
        else:
            return NotImplemented
        # end if
    # end def

    def cdot (self, other):
        """Hadamard matrix product"""
        return self.__class__(*map(operator.mul, self, other))
    # end def

    def normalize_dimension (self, value):
        if value:
            if isinstance(value, (int, float)):
                return value
            else:
                raise ValueError(
                    "Value must be of int or float type, not '{}'."
                    .format(value)
                )
            # end if
        # end if
        return 0
    # end def

    def to_dict (self):
        return dict(x=self.x, y=self.y)
    # end def

    def to_tuple (self):
        return self.xy
    # end def

    @property
    def x (self):
        return self.__x
    # end def

    @x.setter
    def x (self, value):
        self.__x = self.normalize_dimension(value)
    # end def

    @property
    def xy (self):
        """read-only attribute"""
        return (self.x, self.y)
    # end def

    @property
    def y (self):
        return self.__y
    # end def

    @y.setter
    def y (self, value):
        self.__y = self.normalize_dimension(value)
    # end def

# end class


class WorldPoint3D (WorldPoint2D):

    def __init__ (self, x=0, y=0, z=0):
        try:
            (self.x, self.y, self.z) = x
        except:
            (self.x, self.y, self.z) = (x, y, z)
        # end try
    # end def

    def to_dict (self):
        return dict(x=self.x, y=self.y, z=self.z)
    # end def

    def to_tuple (self):
        return self.xyz
    # end def

    @property
    def xyz (self):
        """read-only attribute"""
        return (self.x, self.y, self.z)
    # end def

    @property
    def z (self):
        return self.__z
    # end def

    @z.setter
    def z (self, value):
        self.__z = self.normalize_dimension(value)
    # end def

# end class


class WorldVector2D (WorldPoint2D):

    def __abs__ (self):
        return sum(dim**2 for dim in self)**0.5
    # end def

    def flip (self, query=None):
        query = str(query or "xy").lower()
        if "x" in query:
            self.flip_x()
        # end if
        if "y" in query:
            self.flip_y()
        # end if
        return self
    # end def

    def flip_x (self):
        self.x = -self.x
        return self
    # end def

    def flip_y (self):
        self.y = -self.y
        return self
    # end def

# end class


class WorldVector3D (WorldPoint3D, WorldVector2D):

    def __init__ (self, x=0, y=0, z=0):
        WorldPoint3D.__init__(self, x, y, z)
    # end def

    def flip (self, query=None):
        query = str(query or "xyz").lower()
        if "x" in query:
            self.flip_x()
        # end if
        if "y" in query:
            self.flip_y()
        # end if
        if "z" in query:
            self.flip_z()
        # end if
        return self
    # end def

    def flip_z (self):
        self.z = -self.z
        return self
    # end def

# end class
