#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

try:
    from . import geometry
except:
    import geometry
# end try


class WorldShapeBase2D (geometry.WorldAnchorage2D):

    _OFFSETS = (
        (0, 0), (-1, 0), (0, -1), (-1, -1), (1, -1), (0, 1), (-1, 1),
        (1, 1), (1, 0)
    )

    def __getitem__ (self, key):
        return self.to_tuple()[key]
    # end def

    def __init__ (self, **options):
        super().__init__(**options)
        self.ANCHOR_OFFSETS = dict(
            zip(self.LEGAL_ANCHORS, self._OFFSETS)
        )
    # end def

    def __repr__ (self):
        return str(self.to_tuple())
    # end def

    @property
    def bbox (self):
        """Bounding box - read-only attribute.

        This bounding box is relative to self.origin anchorage point
        and to self.anchor anchorage query."""
        _center = self.center_point
        _hs = self.half_size
        return self.bbox_type(_center - _hs, _center + _hs)
    # end def

    @property
    def bbox_type (self):
        """read-only attribute"""
        return geometry.WorldBoundingBox2D
    # end def

    @property
    def center_point (self):
        """Shape's center point.

        This point is relative to self.origin anchorage point and to
        self.anchor anchorage query."""
        _offset = self.unit_vector.cdot(
            self.ANCHOR_OFFSETS[self.anchor]
        )
        return (self.origin + _offset.cdot(self.half_size))
    # end def

    @property
    def half_size (self):
        """read-only attribute"""
        _w, _h = self.size
        return (_w//2, _h//2)
    # end def

    def points_to_tuple (self, *points):
        return tuple(map(tuple, points))
    # end def

    @property
    def size (self):
        """read-only attribute"""
        # MUST be reimplemented in subclass
        # returns size of shape object
        raise NotImplementedError
    # end def

    def to_tuple (self):
        # MUST be reimplemented in subclass
        # tuple representation of shape object attributes
        raise NotImplementedError
    # end def

# end class


class WorldShapeBase3D (geometry.WorldAnchorage3D, WorldShapeBase2D):

    _OFFSETS = tuple(
        a + (b,)
        for b in (-1, 0, 1)
        for a in WorldShapeBase2D._OFFSETS
    )

    def __init__ (self, **options):
        geometry.WorldAnchorage3D.__init__(self, **options)
        WorldShapeBase2D.__init__(self, **options)
    # end def

    @property
    def bbox_type (self):
        """read-only attribute"""
        return geometry.WorldBoundingBox3D
    # end def

    @property
    def half_size (self):
        """read-only attribute"""
        _w, _h, _d = self.size
        return (_w//2, _h//2, _d//2)
    # end def

# end class


class WorldShapeImage2D (WorldShapeBase2D):

    def __init__ (self, origin, image, **options):
        super().__init__(origin=origin, **options)
        self.image = image
    # end def

    @property
    def height (self):
        """read-only attribute"""
        return self.image.height()
    # end def

    @property
    def size (self):
        """read-only attribute"""
        return (self.width, self.height)
    # end def

    def to_tuple (self):
        return self.bbox.to_tuple()
    # end def

    @property
    def width (self):
        """read-only attribute"""
        return self.image.width()
    # end def

# end class


class WorldShapeCircle2D (WorldShapeBase2D):

    def __init__ (self, origin, radius, **options):
        super().__init__(origin=origin, **options)
        self.radius = radius
    # end def

    @property
    def size (self):
        """read-only attribute"""
        _diameter = 2 * self.radius
        # returns (width, height)
        return (_diameter, _diameter)
    # end def

    def to_tuple (self):
        return (self.center_point, self.radius)
    # end def

# end class


class WorldShapeSphere3D (WorldShapeCircle2D, WorldShapeBase3D):

    def __init__ (self, origin, radius, **options):
        WorldShapeBase3D.__init__(self, **options)
        WorldShapeCircle2D.__init__(self, origin, radius, **options)
    # end def

    @property
    def size (self):
        """read-only attribute"""
        _diameter = 2 * self.radius
        # returns (width, height, depth)
        return (_diameter, _diameter, _diameter)
    # end def

# end class


class WorldShapePolygon2D (WorldShapeBase2D):

    def __init__ (self, origin, *coords, **options):
        super().__init__(origin=origin, **options)
        try:
            self.coords = coords or options.get("coords")
        except:
            raise ValueError(
                "Missing list of coordinates to build "
                "a consistent polygon."
            ) from None
        # end try
    # end def

    @property
    def coords (self):
        return self.__coords
    # end def

    @coords.setter
    def coords (self, values):
        self.__coords = self.points_to_tuple(*values)
    # end def

    @property
    def rel_coords (self):
        """Relative coordinates.

        These coords are relative to self.center_point."""
        _u = self.unit_vector
        _center = self.center_point
        return self.points_to_tuple(
            *(_center + _u.cdot(coords) for coords in self.coords)
        )
    # end def

    @property
    def rel_coords_flat (self):
        """All relative coords are linearized into a single tuple.

        This converts a ((x1, y1), ..., (xn, yn)) tuple of point coords
        to a single (x1, y1, ..., xn, yn) tuple of flat coordinates."""
        return tuple(
            dim for coords in self.rel_coords for dim in coords
        )
    # end def

    @property
    def size (self):
        """read-only attribute"""
        return tuple(
            max(dims) - min(dims) for dims in zip(*self.coords)
        )
    # end def

    def to_tuple (self):
        return self.rel_coords
    # end def

# end class


class WorldShapePolyhedron3D (WorldShapePolygon2D, WorldShapeBase3D):

    def __init__ (self, origin, *coords, **options):
        WorldShapeBase3D.__init__(self, **options)
        WorldShapePolygon2D.__init__(self, origin, *coords, **options)
    # end def

# end class


class WorldShapeRectangle2D (geometry.WorldBoxBase2D, WorldShapeBase2D):

    def __init__ (self, origin, width, height, **options):
        geometry.WorldBoxBase2D.__init__(
            self, width, height, **options
        )
        WorldShapeBase2D.__init__(self, origin=origin, **options)
    # end def

    def to_tuple (self):
        return self.bbox.to_tuple()
    # end def

# end class


class WorldShapeBox3D (geometry.WorldBoxBase3D, WorldShapeBase3D):

    def __init__ (self, origin, width, height, depth, **options):
        geometry.WorldBoxBase3D.__init__(
            self, width, height, depth, **options
        )
        WorldShapeBase3D.__init__(self, origin=origin, **options)
    # end def

    def to_tuple (self):
        return self.bbox.to_tuple()
    # end def

# end class


def run_demo ():
    import tkinter as tk
    _root = tk.Tk()
    _root.title("CuteJim.shapes tkinter demo")
    _root.resizable(0, 0)
    _canvas = tk.Canvas(_root, width=400, height=250, bg="black")
    _canvas.unit_vector = (1, -1)
    def show_shape (shape, shape_factory):
        _cr = (2, 2)
        for _a in shape.LEGAL_ANCHORS:
            _canvas.create_text(
                tuple(shape.origin + (0, -30)),
                text=_a.upper(), fill="white"
            )
            shape.anchor = _a
            shape_factory(shape)
            _origin = shape.origin
            _canvas.create_oval(
                shape.points_to_tuple(_origin - _cr, _origin + _cr),
                fill="cyan"
            )
            shape.origin += (40, 0)
        # end for
    # end def
    _shape = WorldShapeCircle2D(
        origin=(35, 50), radius=10, unit_vector=_canvas.unit_vector,
    )
    _factory = (
        lambda shape:
            _canvas.create_oval(
                shape.bbox.to_flat_tuple(), fill="yellow"
            )
    )
    show_shape(_shape, _factory)
    _shape = WorldShapeRectangle2D(
        origin=(35, 130), width=20, height=10,
        unit_vector=_canvas.unit_vector,
    )
    _factory = (
        lambda shape:
            _canvas.create_rectangle(
                shape.bbox.to_flat_tuple(), fill="orange"
            )
    )
    show_shape(_shape, _factory)
    _shape = WorldShapePolygon2D(
        origin=(35, 210),
        coords=(
            (9.51, 3.09), (-9.51, 3.09), (5.88, -8.09), (0.0, 10.0),
            (-5.88, -8.09)
        ),
        unit_vector=_canvas.unit_vector,
    )
    _factory = (
        lambda shape:
            _canvas.create_polygon(shape.rel_coords, fill="magenta")
    )
    show_shape(_shape, _factory)
    _canvas.pack()
    _root.mainloop()
# end def


if __name__ == "__main__":
    run_demo()
# end if
