#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import tkinter as tk
try:
    from . import game
except:
    import game
# end try


class TkWorld2D (game.World2D):

    def __init__ (self, tk_parent, **options):
        super().__init__(**options)
        self.tk_parent = tk_parent
        self._loop_tid = 0
    # end def

    def end (self, *args, **kw):
        self._running = False
        self.tk_parent.after_cancel(self._loop_tid)
        self.tk_parent.quit()
    # end def

    def game_loop (self, fps, sequence):
        """This is game's main loop.

        Parameters:
            fps: required number of frames per second (steps/second);
            sequence: sorted sequence of world item names to manage;

        Feel free to override this method to best meet your own needs.
        The following code sample is just for example giving."""
        # hook method to reimplement in subclass
        _delay = self.fps_to_delay(fps)
        self.game_loop_after(_delay, sequence)
        self.tk_parent.mainloop()
    # end def

    def game_loop_after (self, delay, sequence):
        self.tk_parent.after_cancel(self._loop_tid)
        if self._running:
            self._loop_tid = self.tk_parent.after(
                int(delay), self.internal_loop, delay, sequence
            )
        # end if
    # end def

    def internal_loop (self, delay, sequence):
        _dt = delay/1000.0
        for _item_name in sequence:
            self.get(_item_name).do_threaded_step(dt=_dt)
        # end for
        self.game_loop_after(delay, sequence)
    # end def

# end class


class TkWorld3D (game.World3D, TkWorld2D):

    def __init__ (self, tk_parent, **options):
        TkWorld2D.__init__(self, tk_parent, **options)
    # end def

# end class


class TkWorldCanvas2D (tk.Canvas):

    DEFAULT_UNIT_VECTOR = (1, -1)

    def __init__ (self, master=None, **options):
        self.unit_vector = options.pop("unit_vector", None)
        super().__init__(master, **options)
    # end def

    @property
    def unit_vector (self):
        return self.__unit_vector
    # end def

    @unit_vector.setter
    def unit_vector (self, value):
        self.__unit_vector = self.vector_type(
            value or self.DEFAULT_UNIT_VECTOR
        )
    # end def

    @property
    def vector_type (self):
        """read-only attribute"""
        return game.WorldVector2D
    # end def

# end class


class TkWorldCanvas3D (TkWorldCanvas2D):

    DEFAULT_UNIT_VECTOR = (1, -1, 1)

    @property
    def vector_type (self):
        """read-only attribute"""
        return game.WorldVector3D
    # end def

# end class


class TkWorldEventManager (game.WorldEventManager):

    pass # please, refer to events.WorldEventManager implementation

# end class


class TkWorldImageManager (game.WorldImageManager):

    def image_factory (self, filepath):
        return tk.PhotoImage(file=filepath)
    # end def

# end class


class TkWorldShapeBase2D:

    DEFAULT_UNIT_VECTOR = (1, -1)

    def __init__ (self, canvas):
        self.canvas = canvas
        # overrides shapes.WorldAnchorage2D.unit_vector attribute
        try:
            self.unit_vector = canvas.unit_vector
        except:
            self.unit_vector = self.DEFAULT_UNIT_VECTOR
        # end try
    # end def

    def finalize (self, *args, **kw):
        self.canvas.delete(self.canvas_id)
    # end def

    def hide (self, *args, **kw):
        self.canvas.itemconfigure(self.canvas_id, state=tk.HIDDEN)
    # end def

    def initialize (self, **tk_options):
        _tkoptions = dict(state=tk.HIDDEN)
        _tkoptions.update(self.options.get("tk_options") or {})
        _tkoptions.update(tk_options)
        self.canvas_id = self.shape_factory(**_tkoptions)
    # end def

    def lower_below (self, tag_or_id):
        self.canvas.tag_lower(self.canvas_id, tag_or_id)
    # end def

    def raise_above (self, tag_or_id):
        self.canvas.tag_raise(self.canvas_id, tag_or_id)
    # end def

    def render (self, new_position=None, **tk_options):
        if new_position:
            self.origin = new_position
        # end if
        _tkoptions = dict(state=tk.NORMAL)
        _tkoptions.update(tk_options)
        self.canvas.coords(self.canvas_id, *self.shape_coords())
        self.canvas.itemconfigure(self.canvas_id, **_tkoptions)
    # end def

    def shape_coords (self, *args, **kw):
        # hook method to reimplement in subclass
        return self.bbox.to_flat_tuple()
    # end def

    def shape_factory (self, **tk_options):
        # hook method to reimplement in subclass
        pass
    # end def

    def show (self, *args, **kw):
        self.canvas.itemconfigure(self.canvas_id, state=tk.NORMAL)
    # end def

# end class


class TkWorldShapeBase3D (TkWorldShapeBase2D):

    DEFAULT_UNIT_VECTOR = (1, -1, 1)

# end class


class TkWorldShapeCircle (TkWorldShapeBase2D, game.WorldShapeCircle2D):

    def __init__ (self, canvas, origin, radius, **options):
        TkWorldShapeBase2D.__init__(self, canvas)
        game.WorldShapeCircle2D.__init__(
            self, origin, radius, **options
        )
    # end def

    def shape_factory (self, **tk_options):
        return self.canvas.create_oval(0, 0, 0, 0, **tk_options)
    # end def

# end class


class TkWorldShapeImage (TkWorldShapeBase2D, game.WorldShapeImage2D):

    def __init__ (self, canvas, origin, image, **options):
        TkWorldShapeBase2D.__init__(self, canvas)
        game.WorldShapeImage2D.__init__(
            self, origin, image, **options
        )
    # end def

    def shape_coords (self, *args, **kw):
        return self.center_point
    # end def

    def shape_factory (self, **tk_options):
        return self.canvas.create_image(
            0, 0, image=self.image, **tk_options
        )
    # end def

# end class


class TkWorldShapeOval (TkWorldShapeBase2D,
                                        game.WorldShapeRectangle2D):

    def __init__ (self, canvas, origin, width, height, **options):
        TkWorldShapeBase2D.__init__(self, canvas)
        game.WorldShapeRectangle2D.__init__(
            self, origin, width, height, **options
        )
    # end def

    def shape_factory (self, **tk_options):
        return self.canvas.create_oval(0, 0, 0, 0, **tk_options)
    # end def

# end class


class TkWorldShapePolygon (TkWorldShapeBase2D,
                                            game.WorldShapePolygon2D):

    def __init__ (self, canvas, origin, *coords, **options):
        TkWorldShapeBase2D.__init__(self, canvas)
        game.WorldShapePolygon2D.__init__(
            self, origin, *coords, **options
        )
    # end def

    def shape_coords (self, *args, **kw):
        return self.rel_coords_flat
    # end def

    def shape_factory (self, **tk_options):
        return self.canvas.create_polygon(0, 0, **tk_options)
    # end def

# end class


class TkWorldShapeRectangle (TkWorldShapeBase2D,
                                            game.WorldShapeRectangle2D):

    def __init__ (self, canvas, origin, width, height, **options):
        TkWorldShapeBase2D.__init__(self, canvas)
        game.WorldShapeRectangle2D.__init__(
            self, origin, width, height, **options
        )
    # end def

    def shape_factory (self, **tk_options):
        return self.canvas.create_rectangle(0, 0, 0, 0, **tk_options)
    # end def

# end class


class TkWorldSoundManager:

    # Caution: tkinter does not manage sounds
    pass

# end class


class TkWorldSprite2D (game.WorldSprite2D):

    DEFAULT_UNIT_VECTOR = (1, -1)

# end class


class TkWorldSprite3D (game.WorldSprite3D):

    DEFAULT_UNIT_VECTOR = (1, -1, 1)

# end class


class TkWorldPlayer2D (TkWorldSprite2D, game.WorldPlayer2D):

    pass # please, refer to visual.WorldPlayer2D implementation

# end class


class TkWorldPlayer3D (TkWorldSprite3D, game.WorldPlayer3D):

    pass # please, refer to visual.WorldPlayer3D implementation

# end class


class TkWorldView (game.WorldView):

    def __init__ (self, canvas, world, name=None, **options):
        super().__init__(world, name, **options)
        self.canvas = canvas
    # end def

# end class


def run_demo ():

    class Player (TkWorldPlayer2D):

        def action_move_down (self, *args, **kw):
            self.position = self.next_step(offset=(0, -1))
        # end def

        def action_move_up (self, *args, **kw):
            self.position = self.next_step(offset=(0, 1))
        # end def

        def do_step (self, *args, **kw):
            pass
        # end def

        def finalize (self):
            pass
        # end def

        def initialize (self):
            self.shape.initialize()
            self.position = self.shape.origin
            self.velocity = (0, 10)
        # end def

        def next_step (self, offset):
            _new_position = (
                self.position + self.velocity.cdot(
                    self.unit_vector.cdot(offset)
                )
            )
            self.shape.origin = _new_position
            _p1, _p2 = self.shape.bbox
            _inbounds = self.world.area.bounds.in_bounds
            if not (_inbounds(_p1) and _inbounds(_p2)):
                self.shape.origin = _new_position = self.position
            # end if
            return _new_position
        # end def

    # end class

    class Ball (TkWorldPlayer2D):

        def do_step (self, *args, **kw):
            self.next_step()
        # end def

        def finalize (self):
            pass
        # end def

        def initialize (self):
            self.shape.initialize()
            self.position = self.shape.origin
            self.velocity = (5, 5)
        # end def

        def next_step (self):
            _new_position = (
                self.position + self.velocity.cdot(self.unit_vector)
            )
            self.shape.origin = self.position = _new_position
            _player = self.world.get("player")
            _collisions = set(
                _player.shape.bbox.entered(*self.shape.bbox)
                + self.world.area.bounds.exited(*self.shape.bbox)
            )
            if _collisions.intersection(("left", "right")):
                self.velocity.flip_x()
            # end if
            if _collisions.intersection(("top", "bottom")):
                self.velocity.flip_y()
            # end if
        # end def

    # end class

    class View (TkWorldView):

        def do_step (self, *args, **kw):
            for _item in self.world.get_items(exclude=self):
                try:
                    _item.shape.render()
                except Exception as e:
                    print("caught exception:", e)
                # end try
            # end for
        # end def

    # end class

    _root = tk.Tk()
    _root.title("Tkinter game demo")
    _root.resizable(0, 0)
    _canvas = TkWorldCanvas2D(_root, background="black")
    _world = TkWorld2D(_root, area=(320, 240))
    _player = Player(_world)
    _player.shape = TkWorldShapeRectangle(
        _canvas, origin=(10, _world.area.center_y), width=10,
        height=30, tk_options=dict(fill="white"),
    )
    _ball = Ball(_world)
    _ball.shape = TkWorldShapeRectangle(
        _canvas, origin=_world.area.center_xy, width=10, height=10,
        tk_options=dict(fill="white"),
    )
    _view = View(_canvas, _world)
    _root.bind("<Up>", _player.action_move_up)
    _root.bind("<Down>", _player.action_move_down)
    _root.bind("<Escape>", _world.end)
    _canvas.configure(**_world.area.to_dict())
    _canvas.pack()
    _world.run()
# end def


if __name__ == "__main__":
    run_demo()
# end if
