#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

try:
    from . import world
    from . import geometry
except:
    import world
    import geometry
# end try


class WorldSprite2D (world.WorldItem):

    DEFAULT_POSITION = None
    DEFAULT_VELOCITY = None
    DEFAULT_UNIT_VECTOR = (1, 1)

    def __init__ (self, world, name=None, **options):
        super().__init__(world, name)
        self.position = options.get("position")
        self.velocity = options.get("velocity")
        self.unit_vector = options.get("unit_vector")
    # end def

    @property
    def point_type (self):
        """read-only attribute"""
        return geometry.WorldPoint2D
    # end def

    @property
    def position (self):
        return self.__position
    # end def

    @position.setter
    def position (self, value):
        self.__position = self.point_type(
            value or self.DEFAULT_POSITION
        )
    # end def

    @property
    def unit_vector (self):
        return self.__unit_vector
    # end def

    @unit_vector.setter
    def unit_vector (self, value):
        self.__unit_vector = self.vector_type(
            value or self.DEFAULT_UNIT_VECTOR
        )
    # end def

    @property
    def vector_type (self):
        """read-only attribute"""
        return geometry.WorldVector2D
    # end def

    @property
    def velocity (self):
        return self.__velocity
    # end def

    @velocity.setter
    def velocity (self, value):
        self.__velocity = self.vector_type(
            value or self.DEFAULT_VELOCITY
        )
    # end def

# end class


class WorldSprite3D (WorldSprite2D):

    DEFAULT_UNIT_VECTOR = (1, 1, 1)

    @property
    def point_type (self):
        """read-only attribute"""
        return geometry.WorldPoint3D
    # end def

    @property
    def vector_type (self):
        """read-only attribute"""
        return geometry.WorldVector3D
    # end def

# end class


class WorldPlayer2D (WorldSprite2D):

    pass # see world.WorldItemInterface for implementation details

# end class


class WorldPlayer3D (WorldSprite3D):

    pass # see world.WorldItemInterface for implementation details

# end class


class WorldView (world.WorldItem):

    pass # see world.WorldItemInterface for implementation details

# end class


def run_demo ():
    class Player (WorldPlayer2D):

        def initialize (self):
            print("initializing", self.name)
            self.position = self.world.area.center_xy
            print(self.name, "position:", self.position)
            self.velocity = (-1, 1)
            print(self.name, "velocity:", self.velocity)
            self.counter = 0
        # end def

        def do_step (self, *args, **kw):
            self.counter += 1
            if self.counter < 10:
                print(
                    self.name, "doing action:", self.counter
                )
                self.position += self.unit_vector.cdot(self.velocity)
            else:
                self.world.end()
            # end if
        # end def

        def finalize (self):
            print("finalizing", self.name)
        # end def

    # end class

    class View (WorldView):

        def do_step (self, *args, **kw):
            print(
                self.name, "showing player's position:",
                self.world.get("player").position
            )
        # end def

        def initialize (self):
            print("initializing", self.name)
        # end def

        def finalize (self):
            print("finalizing", self.name)
        # end def

    # end class

    print("starting demo.")
    _world = world.World2D()
    _player = Player(_world)
    _view = View(_world)
    _world.run(fps=5)
    print("demo ended.")
# end def


if __name__ == "__main__":
    run_demo()
# end if
