#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    The MIT License (MIT)

    Copyright (c) 2015 Raphaël SEBAN

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use, copy,
    modify, merge, publish, distribute, sublicense, and/or sell copies
    of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
    BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
    ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
    CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
"""

import threading
try:
    from . import geometry
except:
    import geometry
# end try


class Conventions:

    def __init__ (self, name=None):
        self.__name = self.naming_convention(name or self.classname())
    # end def

    def classname (self, instance=None):
        return (instance or self).__class__.__name__
    # end def

    @property
    def name (self):
        """read-only attribute"""
        return self.__name
    # end def

    def naming_convention (self, name):
        # here, names are lowercased
        # for case-insensitive indexing convention
        return str(name).lower()
    # end def

# end class


class World2D (Conventions):

    DEFAULT_AREA = (640, 480)

    def __init__ (self, **options):
        super().__init__(name=options.get("name"))
        self.options = options
        self.area = options.get("area")
        self._running = False
        self.__items = dict()
    # end def

    @property
    def area (self):
        return self.__area
    # end def

    @area.setter
    def area (self, value):
        self.__area = self.area_type(value or self.DEFAULT_AREA)
    # end def

    @property
    def area_type (self):
        """read-only attribute"""
        return geometry.WorldArea2D
    # end def

    def end (self, *args, **kw):
        self._running = False
    # end def

    def fps_to_delay (self, fps):
        # in milliseconds
        return 1000 // fps
    # end def

    def game_loop (self, fps, sequence):
        """This is game's main loop.

        Parameters:
            fps: required number of frames per second (steps/second);
            sequence: sorted sequence of world item names to manage;

        Feel free to override this method to best meet your own needs.
        The following code sample is just for example giving."""
        # hook method to reimplement in subclass
        _delay = self.fps_to_delay(fps)
        _dt = _delay/1000.0
        while self._running:
            for _item_name in sequence:
                self.get(_item_name).do_threaded_step(dt=_dt)
            # end for
            self.sleep(_dt)
        # end while
    # end def

    def get (self, item_name, strict=True):
        item_name = self.naming_convention(item_name)
        if strict:
            return self.items[item_name]
        else:
            return self.items.get(item_name)
        # end if
    # end def

    def get_items (self, exclude=None):
        _items = set(self.items.values())
        try:
            _items.difference_update(exclude)
        except:
            _items.discard(exclude)
        # end try
        return _items
    # end def

    @property
    def items (self):
        """read-only attribute"""
        return self.__items
    # end def

    def register (self, world_item, name=None):
        name = self.naming_convention(name or world_item.name)
        if name in self.items:
            raise KeyError(
                "Item name '{}' is already registered.".format(name)
            )
        else:
            self.items[name] = world_item
            world_item.registered_as_name = name
        # end if
    # end def

    def run (self, sequence=None, fps=30):
        sequence = tuple(sequence or sorted(self.items.keys()))
        for _item_name in sequence:
            self.get(_item_name).initialize()
        # end for
        self._running = True
        self.game_loop(fps, sequence)
        for _item_name in sequence:
            _item = self.get(_item_name)
            _item.finalize()
            _item.leave_world()
        # end for
        self.wait()
    # end def

    def sleep (self, delay):
        # hook method
        import time
        time.sleep(delay)
    # end def

    def unregister (self, item):
        if isinstance(item, str):
            self.items.pop(self.naming_convention(item), None)
        else:
            # remove ALL occurrences of item
            for _name, _item in self.items.copy().items():
                if _item is item:
                    self.items.pop(_name, None)
                # end if
            # end for
        # end if
    # end def

    def wait (self):
        # wait until all world items are gone (unregistered)
        while self.items:
            self.wait_do_events()
        # end while
    # end def

    def wait_do_events (self, *args, **kw):
        # hook method
        pass
    # end def

# end class


class World3D (World2D):

    DEFAULT_AREA = (1000, 1000, 1000)

    @property
    def area_type (self):
        """read-only attribute"""
        return geometry.WorldArea3D
    # end def

# end class


class WorldItemInterface:
    """Gathers all hook methods to be reimplemented in subclasses."""

    def do_step (self, *args, **kw):
        pass
    # end def

    def finalize (self):
        pass
    # end def

    def initialize (self):
        pass
    # end def

# end class


class WorldItem (Conventions, WorldItemInterface):

    def __init__ (self, world, name=None, **options):
        Conventions.__init__(self, name)
        self.thread = None
        self.options = options
        self.world = world
        self.world.register(self)
    # end def

    def do_threaded_step (self, *args, **kw):
        if self.thread and self.thread.is_alive():
            raise RuntimeError(
                "World item '{}' takes too long to execute one step."
                .format(self.classname())
            )
        else:
            self.thread = threading.Thread(
                target=self.do_step, args=args, kwargs=kw, daemon=True
            )
            self.thread.start()
        # end if
    # end def

    def enter_world (self, world=None):
        world = world or self.world
        world.register(self)
    # end def

    def leave_world (self, world=None):
        world = world or self.world
        world.unregister(self)
    # end def

    def move_to_world (self, world):
        self.leave_world(self.world)
        self.world = world
        self.enter_world(self.world)
    # end def

# end class


def run_demo ():

    class Player (WorldItem):

        def initialize (self):
            self.counter = 0
        # end def

        def do_step (self, *args, **kw):
            if self.counter < 10:
                self.counter += 1
                print("counter:", self.counter)
            else:
                self.world.end()
            # end if
        # end def

    # end class

    print("starting demo.")
    _world = World2D()
    _player = Player(_world)
    _world.run(fps=5)
    print("demo ended.")
# end def


if __name__ == "__main__":
    run_demo()
# end if
